"""A library of functions for enforcing and enacting the documentation rules.

THE RULES
You cannot write to a CirrusVolume unless you've passed it:
(1) A set of sources from which this was created (e.g., another CloudVolume
path or a freeform justification like "tracer annotation"). This must be
formatted as a list[str], and any sources that haven't been logged previously
will be added to the current sources field of the provenance file

(2) The motivation for creating the volume (str) has been logged and you've
included that motivation when instantiating the object. A volume can have
multiple motivation notes, and your motivation only need match one of them.

(3) You've passed a Process (a code environment & parameters) to the class.
The process will be logged unless another process with the same task
description has already been logged. These are defined by provenance-tools.
"""
from __future__ import annotations

import warnings
from typing import Optional

import cloudvolume as cv
import provenancetoolbox as ptb


def check_writing_rules(sources: Optional[list[str]] = None,
                        motivation: Optional[str] = None,
                        process: Optional[ptb.Process] = None
                        ) -> None:
    """Checks the rules for write-access to a CloudVolume.

    Checks whether the provided fields are sufficient to allow writing
    to a volume.

    Args:
        sources: A list of descriptions for where this volume came from.
        motivation: The motivation for creating this volume.
        process: The process that describes the volume's creation.

    Raises:
        AssertionError: One of the write rules has been violated.
    """
    assert all(v is not None for v in [sources, motivation, process]
               ), (""
                   "Need to define sources, motivation and process in order"
                   " to write to this volume")

    # Checking sources
    assert isinstance(sources, list)
    assert all(isinstance(v, str) for v in sources)

    # Checking motivation
    assert isinstance(motivation, str)

    # Checking process
    assert isinstance(process, ptb.Process)


def add_sources(cloudvolume: cv.CloudVolume,
                sources: Optional[list[str]] = None
                ) -> None:
    """Logs sources in a CloudVolume if they don't already exist.

    Args:
        cloudvolume: A CloudVolume.
        sources: A list of cloudpaths or freeform sources.
    """
    currentsources = set(cloudvolume.provenance.sources)

    newsources = currentsources.union(sources)

    if len(newsources) > len(currentsources):
        cloudvolume.provenance.sources = list(newsources)
        cloudvolume.commit_provenance()


def add_motivation(cloudvolume: cv.CloudVolume,
                   motivation: Optional[str] = None
                   ) -> None:
    """Logs a motivation in a CloudVolume if it doesn't already exist.

    Args:
        cloudvolume: A CloudVolume.
        motivation: A reason why this CloudVolume was created.
    """
    if ptb.note_absent(cloudvolume, motivation, ptb.NoteType.MOTIVATION):
        ptb.add_motivation(cloudvolume, motivation)


def add_process(cloudvolume: cv.CloudVolume,
                process: Optional[ptb.Process] = None
                ) -> None:
    
    """Logs a ptb.Process in a CloudVolume if not already logged.

    Args:
        cloudvolume: A CloudVolume.
        process: A processing step.
    """
    if ptb.process_absent(cloudvolume, process):
        ptb.log_process(cloudvolume, process)
    else:
        warnings.warn('Process with the same description already logged.'
                      ' Skipping')


def documentvolume(cloudvolume: cv.CloudVolume,
                   sources: Optional[list[str]] = None,
                   motivation: Optional[str] = None,
                   process: Optional[ptb.Process] = None
                   ) -> None:
    """Logs all required fields for writing to a CirrusVolume.

    Args:
        cloudvolume: A CloudVolume.
        sources: A list of cloudpaths or freeform sources.
        motivation: A reason why this CloudVolume was created.
        process: A processing step.
    """
    add_sources(cloudvolume, sources)
    add_motivation(cloudvolume, motivation)
    add_process(cloudvolume, process)



"""A wrapper for the generic CloudVolume class.

Returns a separate wrapper for each registered CloudVolume front end.

Typical usage:
    import cirrusvolume as cv

    cv.CloudVolume(cloudpath, ...)
"""
from __future__ import annotations

from typing import Optional, Type

import cloudvolume as cv
import provenancetoolbox as ptb

from . import rules


REGISTERED_PLUGINS = dict()


def register_plugin(key: Type[cv.CloudVolume],
                    creation_function:
                        Callable[[Type[cv.CloudVolume]],
                                 Callable[[cv.CloudVolume], CloudVolume]]
                    ) -> None:
    """Connects CloudVolume classes to CirrusVolume wrappers.

    Args:
        key: A type of CloudVolume.
        creation_function: A CirrusVolume wrapper class.
    """
    REGISTERED_PLUGINS[key] = creation_function


class CloudVolume(cv.CloudVolume):
    """A CirrusVolume wrapper class for CloudVolumes.

    See cloudvolume.CloudVolume for original functionality. The CirrusVolume
    wrapper enforces rules for documentation before it allows data writing or
    modification.

    Additional attributes:
        sources: A list of cloudpaths or freeform sources.
        motivation: A reason why this CloudVolume was created.
        process: A processing step.
    """
    def __new__(cls,
                *args,
                sources: Optional[list[str]] = None,
                motivation: Optional[str] = None,
                process: Optional[ptb.Process] = None,
                **kwargs):
        cloudvolume = cv.CloudVolume(*args, **kwargs)

        return REGISTERED_PLUGINS[type(cloudvolume)](
                   cloudvolume, sources, motivation, process)

    # need to re-define this here bc it automatically writes
    @classmethod
    def from_numpy(cls,
                   *args,
                   sources: Optional[list[str]] = None,
                   motivation: Optional[str] = None,
                   process: ptb.Process = None,
                   **kwargs):
        """Create a new dataset from a numpy array.

        See cloudvolume.Cloudvolume for more information. Also enforces
        documentation rules before writing data.

        Additional args:
            sources: A list of cloudpaths or freeform sources.
            motivation: A reason why this CloudVolume was created.
            process: A processing step.
        """
        rules.check_writing_rules(sources, motivation, process)

        cloudvolume = super().from_numpy(cls, *args, **kwargs)

        rules.add_sources(cloudvolume, sources)
        rules.add_motivation(cloudvolume, motivation)
        rules.add_process(cloudvolume, process)

        return REGISTERED_PLUGINS[type(cloudvolume)](
                   cloudvolume, sources, motivation, process)



import sys
import setuptools
from setuptools import setup, find_packages


__version__ = '0.1.3'


setup(
    name='cirrus-volume',
    version=__version__,
    description='Making happy (well-documented) CloudVolumes',
    author='Nicholas Turner',
    author_email='nturner@zetta.ai',
    url='https://github.com/ZettaAI/cirrus-volume',
    packages=setuptools.find_packages(),
    install_requires=['provenance-toolbox', 'cloud-volume==8.2.0']
)



"""A wrapper for CloudVolumePrecomputed."""
from __future__ import annotations

from typing import Optional

import cloudvolume as cv
from cloudvolume.frontends.precomputed import CloudVolumePrecomputed
import provenancetoolbox as ptb

from . import rules
from .volume import register_plugin


def register():
    """Registration function for CirrusVolumePrecomputed.

    See volume.register_plugin for more information.
    """
    register_plugin(CloudVolumePrecomputed, CirrusVolumePrecomputed)


class CirrusVolumePrecomputed(CloudVolumePrecomputed):

    def __init__(self,
                 cloudvolume: cv.CloudVolume,
                 sources: Optional[list[str]] = None,
                 motivation: Optional[str] = None,
                 process: Optional[ptb.Process] = None
                 ):
        # Copying the CloudVolume attributes
        self.config = cloudvolume.config
        self.cache = cloudvolume.cache
        self.meta = cloudvolume.meta

        self.image = cloudvolume.image
        self.mesh = cloudvolume.mesh
        self.skeleton = cloudvolume.skeleton

        self.green_threads = cloudvolume.green_threads

        self.mip = cloudvolume.mip
        self.pid = cloudvolume.pid

        # CirrusVolume-specific atrributes
        self.sources = sources
        self.motivation = motivation
        self.process = process

    # Overriding all methods that allow writing to the image
    # with versions that check the rules first
    def __setitem__(self, slices, img):
        rules.check_writing_rules(self.sources, self.motivation, self.process)

        rules.documentvolume(self, self.sources,
                             self.motivation, self.process)

        super().__setitem__(self, slices, img)

    def upload_from_shared_memory(self,
                                  location,
                                  bbox,
                                  order='F',
                                  cutout_bbox=None) -> None:
        rules.check_writing_rules(self.sources, self.motivation, self.process)

        rules.documentvolume(self, self.sources,
                             self.motivation, self.process)

        super().upload_from_shared_memory(self, location, bbox,
                                          order=order, cutout_bbox=cutout_bbox)

    def upload_from_file(self,
                         location, bbox, order='F', cutout_bbox=None) -> None:
        rules.check_writing_rules(self.sources, self.motivation, self.process)

        rules.documentvolume(self, self.sources,
                             self.motivation, self.process)

        super().upload_from_file(self, location, bbox,
                                 order=order, cutout_bbox=cutout_bbox)
